# SPDX-FileCopyrightText: 2024 Helmholtz-Zentrum hereon GmbH
#
# SPDX-License-Identifier: LGPL-3.0-only

# coding: utf-8
"""Test file for building the face_edge_connectivity"""
from typing import Callable, List

import numpy as np

from psy_ugrid import ugrid


def test_create_dual_node_mesh(
    same_upon_permutation: Callable[[List, List], bool]
):
    r"""Test for a mesh like

     /|\
    /_|_\
    """
    nodex = [0, 1, 1, 2]
    nodey = [0, 1, 0, 0]
    faces = [[0, 1, 2], [2, 1, 3]]

    edges = [[0, 1], [1, 2], [2, 0], [1, 3], [2, 3]]
    grid = ugrid.UGrid(
        node_lon=nodex, node_lat=nodey, faces=faces, edges=edges
    )

    ref = [
        [4, 6, 0, 7, -999],
        [6, 4, 5, 8, 1],
        [5, 4, 7, 2, 9],
        [8, 5, 9, 3, -999],
    ]

    dual_faces, dual_nodes = grid._create_dual_node_mesh()

    # check length of nodes.
    # 4 original nodes + 2 face centers + 4 edge centers
    # (edge 1-2 does not appear)
    assert len(dual_nodes) == 10

    # check faces
    assert np.ma.isMA(dual_faces)
    test = dual_faces.filled(-999).tolist()
    tests = [same_upon_permutation(t, r) for t, r in zip(test, ref)]
    try:
        assert all(tests)
    except AssertionError:
        # for better error message
        assert test == ref


def test_create_dual_node_mesh_na(
    same_upon_permutation: Callable[[List, List], bool]
):
    r"""Test for a mesh like

    |‾|\
    |_|_\
    """
    nodex = [0, 0, 1, 1, 2]
    nodey = [0, 1, 1, 0, 0]
    faces = np.ma.array(
        [[0, 1, 2, 3], [2, 4, 3, -999]],
        mask=[[False, False, False, False], [False, False, False, True]],
    )

    edges = [[0, 1], [1, 2], [2, 3], [3, 0], [2, 4], [4, 3]]
    grid = ugrid.UGrid(
        node_lon=nodex, node_lat=nodey, faces=faces, edges=edges
    )

    ref = [
        [5, 7, 0, 9, -999],
        [7, 5, 8, 1, -999],
        [8, 5, 6, 10, 2],
        [6, 5, 9, 3, 11],
        [10, 6, 11, 4, -999],
    ]

    dual_faces, dual_nodes = grid._create_dual_node_mesh()

    # check length of nodes.
    # 5 original nodes + 2 face centers + 5 edge centers
    # (edge 2-3 does not appear)
    assert len(dual_nodes) == 12

    # check faces
    assert np.ma.isMA(dual_faces)
    test = dual_faces.filled(-999).tolist()
    tests = [same_upon_permutation(t, r) for t, r in zip(test, ref)]
    try:
        assert all(tests)
    except AssertionError:
        # for better error message
        assert test == ref



"""psy-ugrid decoder for UGRID conventions

This module defines the decoder that we use to decode the UGRID conventions.
It is supposed to overwrite the one defined in the :mod:`psyplot.data` module
and has been originally developed in

https://codebase.helmholtz.cloud/psyplot/psyplot/-/merge_requests/31
"""

# SPDX-FileCopyrightText: 2024 Helmholtz-Zentrum hereon GmbH
#
# SPDX-License-Identifier: LGPL-3.0-only


from __future__ import annotations

from typing import TYPE_CHECKING, Dict, List
from warnings import warn

import numpy as np
import psyplot.data as psyd
import xarray as xr
from psyplot.docstring import docstrings

if TYPE_CHECKING:
    from psy_ugrid.ugrid import UGrid


xr_version = tuple(map(int, xr.__version__.split(".")[:2]))


class UGridDecoder(psyd.CFDecoder):
    """Decoder for UGrid data sets."""

    #: mapping from grid name to the :class:`gridded.pyugrid.ugrid.UGrid`
    # object representing it
    _grids: Dict[str, UGrid] = {}

    #: True if the data of the CFDecoder supports the extraction of a subset of
    #: the data based on the indices.
    #:
    #: For UGRID conventions, this is not easily possible because the
    #: extraction of a subset breaks the connectivity information of the mesh
    supports_spatial_slicing: bool = False

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._grids = {}

    def clear_cache(self):
        """Clear the cache and remove the UGRID instances."""
        self._grids.clear()

    def is_unstructured(self, *args, **kwargs):
        """Reimpletemented to return always True. Any ``*args`` and ``**kwargs``
        are ignored"""
        return True

    @docstrings.get_sections(base="UGridDecoder.get_mesh")
    def get_mesh(self, var, coords=None):
        """Get the mesh variable for the given `var`

        Parameters
        ----------
        var: xarray.Variable
            The data source whith the ``'mesh'`` attribute
        coords: dict
            The coordinates to use. If None, the coordinates of the dataset of
            this decoder is used

        Returns
        -------
        xarray.Coordinate
            The mesh coordinate"""
        mesh = var.attrs.get("mesh")
        if mesh is None:
            return None
        if coords is None:
            coords = self.ds.coords
        return coords.get(mesh, self.ds.coords.get(mesh))

    @docstrings.with_indent(8)
    def get_ugrid(self, var, coords=None, loc="infer"):
        """Get the :class:`~gridded.pyugrid.ugrid.UGrid` mesh object.

        This method creates a :class:`gridded.pyugrid.ugrid.UGrid` object for
        a given variable, depending on the corresponding ``'mesh'`` attribute.

        Parameters
        ----------
        %(UGridDecoder.get_mesh.parameters)s
        dual: {"infer", "node", "edge", "face"}
            If "node" or "edge", the dual grid will be computed.

        Returns
        -------
        gridded.pyugrid.ugrid.UGrid
            The UGrid object representing the mesh.
        """
        from psy_ugrid.ugrid import UGrid

        def get_coord(cname, raise_error=True):
            try:
                ret = coords[cname]
            except KeyError:
                if cname not in self.ds.coords:
                    if raise_error:
                        raise
                    return None
                else:
                    ret = self.ds.coords[cname]
                    try:
                        idims = var.psy.idims
                    except AttributeError:  # got xarray.Variable
                        idims = {}
                    ret = ret.isel(
                        **{d: sl for d, sl in idims.items() if d in ret.dims}
                    )
            if "start_index" in ret.attrs:
                return ret - int(ret.start_index)
            else:
                return ret

        mesh = self.get_mesh(var, coords)

        if not mesh:
            mesh = self.get_mesh(var)

        if not mesh:
            raise ValueError(
                "Could not find the mesh variable in the coordinates."
            )

        if mesh.name in self._grids:
            grid = self._grids[mesh.name]
        else:
            required_parameters = ["faces"]

            parameters = {
                "faces": "face_node_connectivity",
                "face_face_connectivity": "face_face_connectivity",
                "edges": "edge_node_connectivity",
                "boundaries": "boundary_node_connectivity",
            }

            coord_parameters = {
                "face_coordinates": "face_coordinates",
                "edge_coordinates": "edge_coordinates",
                "boundary_coordinates": "boundary_coordinates",
            }

            x_nodes, y_nodes = self.get_nodes(mesh, coords)

            kws = {
                "node_lon": x_nodes,
                "node_lat": y_nodes,
                "mesh_name": mesh.name,
            }

            coords = dict(coords or {})

            # check for face_dimension and edge_dimension and make sure they
            # appear last in the list
            if "face_dimension" in mesh.attrs:
                cname = mesh.attrs["face_node_connectivity"]
                faces = get_coord(cname)
                coords[cname] = faces.transpose(
                    mesh.attrs["face_dimension"], ...
                )

            if (
                "edge_dimension" in mesh.attrs
                and "edge_node_connectivity" in mesh.attrs
            ):
                cname = mesh.attrs["edge_node_connectivity"]
                edges = get_coord(cname)
                coords[cname] = edges.transpose(
                    mesh.attrs["edge_dimension"], ...
                )

            if (
                "edge_dimension" in mesh.attrs
                and "edge_face_connectivity" in mesh.attrs
            ):
                cname = mesh.attrs["edge_face_connectivity"]
                ef_conn = get_coord(cname)
                coords[cname] = ef_conn.transpose(
                    mesh.attrs["edge_dimension"], ...
                )

            for key, attr in parameters.items():
                if attr in mesh.attrs:
                    coord = get_coord(
                        mesh.attrs[attr], key in required_parameters
                    )
                    if coord is not None:
                        kws[key] = coord.values

            for key, attr in coord_parameters.items():
                if attr in mesh.attrs:
                    xname, yname = mesh.attrs[attr].split()
                    kws[key] = np.dstack([get_coord(xname), get_coord(yname)])[
                        0
                    ]

            # now we have to turn NaN into masked integer arrays
            for param in parameters:
                if kws.get(param) is not None:
                    arr = kws[param]
                    mask = np.isnan(arr)
                    if mask.any():
                        arr = np.where(mask, -999, arr).astype(int)
                    kws[param] = np.ma.masked_where(mask, arr)

            grid = UGrid(**kws)
            self._grids[mesh.name] = grid

        # create the dual mesh if necessary
        if loc == "infer":
            loc = self.infer_location(var, coords, grid)

        if loc in ["node", "edge"]:
            dual_name = grid.mesh_name + "_dual_" + loc
            if dual_name in self._grids:
                grid = self._grids[dual_name]
            else:
                grid = grid.create_dual_mesh(loc)
                grid.mesh_name = dual_name
                self._grids[dual_name] = grid

        return grid

    @classmethod
    @docstrings.dedent
    def can_decode(cls, ds, var):
        """
        Check whether the given variable can be decoded.

        Returns True if a mesh coordinate could be found via the
        :meth:`get_mesh` method

        Parameters
        ----------
        %(CFDecoder.can_decode.parameters)s

        Returns
        -------
        %(CFDecoder.can_decode.returns)s"""
        return cls(ds).get_mesh(var) is not None

    @docstrings.dedent
    def get_triangles(
        self,
        var,
        coords=None,
        convert_radian=True,
        copy=False,
        src_crs=None,
        target_crs=None,
        nans=None,
        stacklevel=1,
    ):
        """
        Get the of the given coordinate.

        Parameters
        ----------
        %(CFDecoder.get_triangles.parameters)s

        Returns
        -------
        %(CFDecoder.get_triangles.returns)s

        Notes
        -----
        If the ``'location'`` attribute is set to ``'node'``, a delaunay
        triangulation is performed using the
        :class:`matplotlib.tri.Triangulation` class.

        .. todo::
            Implement the visualization for UGrid data shown on the edge of the
            triangles"""
        warn(
            "The 'get_triangles' method is depreceated and will be removed "
            "soon! Use the 'get_cell_node_coord' method!",
            DeprecationWarning,
            stacklevel=stacklevel,
        )
        from matplotlib.tri import Triangulation

        if coords is None:
            coords = self.ds.coords

        def get_coord(coord):
            return coords.get(coord, self.ds.coords.get(coord))

        mesh = self.get_mesh(var, coords)
        nodes = self.get_nodes(mesh, coords)
        if any(n is None for n in nodes):
            raise ValueError("Could not find the nodes variables!")
        xvert, yvert = nodes
        xvert = xvert.values
        yvert = yvert.values
        loc = var.attrs.get("location", "face")
        if loc == "face":
            triangles = get_coord(
                mesh.attrs.get("face_node_connectivity", "")
            ).values
            if triangles is None:
                raise ValueError(
                    "Could not find the connectivity information!"
                )
        elif loc == "node":
            triangles = None
        else:
            raise ValueError(
                "Could not interprete location attribute (%s) of mesh "
                "variable %s!" % (loc, mesh.name)
            )

        if convert_radian:
            for coord in nodes:
                if coord.attrs.get("units") == "radian":
                    coord = coord * 180.0 / np.pi
        if src_crs is not None and src_crs != target_crs:
            if target_crs is None:
                raise ValueError(
                    "Found %s for the source crs but got None for the "
                    "target_crs!" % (src_crs,)
                )
            xvert = xvert[triangles].ravel()
            yvert = yvert[triangles].ravel()
            arr = target_crs.transform_points(src_crs, xvert, yvert)
            xvert = arr[:, 0]
            yvert = arr[:, 1]
            if loc == "face":
                triangles = np.reshape(range(len(xvert)), (len(xvert) // 3, 3))

        return Triangulation(xvert, yvert, triangles)

    @docstrings.dedent
    def get_cell_node_coord(self, var, coords=None, axis="x", nans=None):
        """
        Checks whether the bounds in the variable attribute are triangular

        Parameters
        ----------
        %(CFDecoder.get_cell_node_coord.parameters)s

        Returns
        -------
        %(CFDecoder.get_cell_node_coord.returns)s"""
        if coords is None:
            coords = self.ds.coords

        idims = self.get_coord_idims(coords)

        def get_coord(coord):
            coord = coords.get(coord, self.ds.coords.get(coord))
            return coord.isel(
                **{d: sl for d, sl in idims.items() if d in coord.dims}
            )

        mesh = self.get_mesh(var, coords)
        if mesh is None:
            return
        nodes = self.get_nodes(mesh, coords)
        if not len(nodes):
            raise ValueError(
                "Could not find the nodes variables for the %s "
                "coordinate!" % axis
            )
        vert = nodes[0 if axis == "x" else 1]
        if vert is None:
            raise ValueError(
                "Could not find the nodes variables for the %s "
                "coordinate!" % axis
            )

        grid = self.get_ugrid(var, coords)

        faces = grid.faces
        if np.ma.isMA(faces) and faces.mask.any():
            isnull = faces.mask
            faces = faces.filled(-999).astype(int)
            for i in range(faces.shape[1]):
                mask = isnull[:, i]
                if mask.any():
                    for j in range(i, faces.shape[1]):
                        faces[mask, j] = faces[mask, j - i]

        node = grid.nodes[..., 0 if axis == "x" else 1]
        bounds = node[faces]

        loc = self.infer_location(var, coords)

        dim0 = "__face" if loc == "node" else var.dims[-1]
        return xr.DataArray(
            bounds,
            coords={
                key: val for key, val in coords.items() if (dim0,) == val.dims
            },
            dims=(
                dim0,
                "__bnds",
            ),
            name=vert.name + "_bnds",
            attrs=vert.attrs.copy(),
        )

    @docstrings.with_indent(8)
    def infer_location(self, var, coords=None, grid=None):
        """Infer the location for the variable.

        Parameters
        ----------
        %(UGridDecoder.get_mesh.parameters)s
        grid: gridded.pyugrid.ugrid.UGrid
            The grid for this variable. If None, it will be created using the
            :meth:`get_ugrid` method (if necessary)

        Returns
        -------
        str
            ``"node"``, ``"face"`` or ``"edge"``
        """
        if coords is None:
            coords = self.ds.coords
        if not var.attrs.get("location"):
            if grid is None:
                grid = self.get_ugrid(var, coords, loc="face")
            # get the axis of the spatial dimension by looking up possible
            # dimension names. We cannot use get_xname here because this uses
            # infer_location
            mesh = self.get_mesh(var, coords)
            possible_dims = set()
            for attr in [
                "node_coordinates",
                "edge_coordinates",
                "face_coordinates",
            ]:
                if attr in mesh.attrs:
                    for vname in mesh.attrs[attr].split():
                        cvar = coords.get(vname)
                        if cvar is not None:
                            possible_dims.update(cvar.dims)
            for attr in [
                "face_node_connectivity",
                "edge_node_connectivity",
                "face_edge_connectivity",
                "face_face_connectivity",
                "edge_face_connectivity",
            ]:
                vname = mesh.attrs.get(attr)
                if vname:
                    cvar = coords.get(vname)
                    if cvar is not None:
                        possible_dims.update(cvar.dims)
            found = possible_dims.intersection(var.dims)
            if found and len(found) == 1:
                axis = var.dims.index(next(iter(found)))
            else:
                axis = -1
            loc = grid.infer_location(var, axis)
        else:
            loc = var.attrs["location"]
        return loc

    @staticmethod
    @docstrings.dedent
    def decode_coords(ds, gridfile=None):
        """
        Reimplemented to set the mesh variables as coordinates

        Parameters
        ----------
        %(CFDecoder.decode_coords.parameters)s

        Returns
        -------
        %(CFDecoder.decode_coords.returns)s"""
        extra_coords = set(ds.coords)
        if gridfile is not None and not isinstance(gridfile, xr.Dataset):
            gridfile = psyd.open_dataset(gridfile)
        for var in ds.variables.values():
            if "mesh" in var.attrs:
                mesh = var.attrs["mesh"]
                if mesh not in extra_coords:
                    extra_coords.add(mesh)
                    try:
                        mesh_var = ds.variables[mesh]
                    except KeyError:
                        if gridfile is not None:
                            try:
                                mesh_var = gridfile.variables[mesh]
                            except KeyError:
                                warn("Could not find mesh variable %s" % mesh)
                                continue
                        else:
                            warn("Could not find mesh variable %s" % mesh)
                            continue

                    parameters = [
                        "node_coordinates",
                        "face_node_connectivity",
                        "face_face_connectivity",
                        "edge_node_connectivity",
                        "edge_face_connectivity",
                        "boundary_node_connectivity",
                        "face_coordinates",
                        "edge_coordinates",
                        "boundary_coordinates",
                    ]

                    for param in parameters:
                        if param in mesh_var.attrs:
                            extra_coords.update(mesh_var.attrs[param].split())
        if gridfile is not None:
            ds.update(
                {
                    k: v
                    for k, v in gridfile.variables.items()
                    if k in extra_coords
                }
            )
        if xr_version < (0, 11):
            ds.set_coords(
                extra_coords.intersection(ds.variables), inplace=True
            )
        else:
            ds._coord_names.update(extra_coords.intersection(ds.variables))
        return ds

    def get_nodes(self, coord, coords=None):
        """Get the variables containing the definition of the nodes

        Parameters
        ----------
        coord: xarray.Coordinate
            The mesh variable
        coords: dict, optional
            The coordinates to use to get node coordinates"""
        if coords is None:
            coords = {}

        def get_coord(coord):
            return coords.get(coord, self.ds.coords.get(coord))

        return list(
            map(get_coord, coord.attrs.get("node_coordinates", "").split()[:2])
        )

    @docstrings.with_indent(8)
    def get_xname(self, var, coords=None):
        """Get the name of the spatial dimension

        Parameters
        ----------
        %(CFDecoder.get_y.parameters)s

        Returns
        -------
        str
            The dimension name
        """

        def get_dim(name):
            coord = coords.get(
                name, ds.coords.get(name, ds.variables.get(name))
            )
            if coord is None:
                raise KeyError(f"Missing {loc} coordinate {name}")
            else:
                return coord.dims[0]

        ds = self.ds
        loc = self.infer_location(var, coords)
        mesh = self.get_mesh(var, coords)
        coords = coords or ds.coords
        if loc == "node":
            return get_dim(mesh.node_coordinates.split()[0])
        elif loc == "edge":
            return get_dim(mesh.edge_node_connectivity)
        else:
            return get_dim(mesh.face_node_connectivity)

    @docstrings.with_indent(8)
    def get_yname(self, var, coords=None):
        """Get the name of the spatial dimension

        Parameters
        ----------
        %(CFDecoder.get_y.parameters)s

        Returns
        -------
        str
            The dimension name
        """
        return self.get_xname(var, coords)  # x- and y-dimensions are the same

    @docstrings.with_indent(8)
    def get_zname(self, var, coords=None):
        """Get the name of the vertical dimension

        Parameters
        ----------
        %(CFDecoder.get_y.parameters)s

        Returns
        -------
        str
            The dimension name
        """
        # reimplement to make sure we do not interfere with x- and y-dimension

        dim = super().get_zname(var, coords)

        t_or_x = [self.get_tname(var, coords), self.get_xname(var, coords)]

        if dim and var.dims and dim in t_or_x:
            # wrong dimension has been chosen by the default approach
            for dim in var.dims:
                if dim not in t_or_x:
                    break
        return dim

    @docstrings.dedent
    def get_x(self, var, coords=None):
        """
        Get the centers of the triangles in the x-dimension

        Parameters
        ----------
        %(CFDecoder.get_y.parameters)s

        Returns
        -------
        %(CFDecoder.get_y.returns)s"""
        if coords is None:
            coords = self.ds.coords
        # first we try the super class
        ret = super(UGridDecoder, self).get_x(var, coords)
        # but if that doesn't work because we get the variable name in the
        # dimension of `var`, we use the means of the faces
        if (
            ret is None
            or ret.name in var.dims
            or (hasattr(var, "mesh") and ret.name == var.mesh)
        ):
            loc = self.infer_location(var, coords)
            x = self.get_nodes(self.get_mesh(var, coords), coords)[0]
            if loc == "node":
                return x
            else:
                grid = self.get_ugrid(var, coords, loc)
                if grid.face_coordinates is None:
                    grid.build_face_coordinates()
                try:
                    cls = xr.IndexVariable
                except AttributeError:  # xarray < 0.9
                    cls = xr.Coordinate
                return cls(
                    x.name, grid.face_coordinates[..., 0], attrs=x.attrs.copy()
                )
        else:
            return ret

    @docstrings.dedent
    def get_y(self, var, coords=None):
        """
        Get the centers of the triangles in the y-dimension

        Parameters
        ----------
        %(CFDecoder.get_y.parameters)s

        Returns
        -------
        %(CFDecoder.get_y.returns)s"""
        if coords is None:
            coords = self.ds.coords
        # first we try the super class
        ret = super(UGridDecoder, self).get_y(var, coords)
        # but if that doesn't work because we get the variable name in the
        # dimension of `var`, we use the means of the triangles
        if (
            ret is None
            or ret.name in var.dims
            or (hasattr(var, "mesh") and ret.name == var.mesh)
        ):
            loc = self.infer_location(var, coords)
            y = self.get_nodes(self.get_mesh(var, coords), coords)[1]
            if loc == "node":
                return y
            else:
                grid = self.get_ugrid(var, coords, loc)
                if grid.face_coordinates is None:
                    grid.build_face_coordinates()
                try:
                    cls = xr.IndexVariable
                except AttributeError:  # xarray < 0.9
                    cls = xr.Coordinate
                return cls(
                    y.name, grid.face_coordinates[..., 1], attrs=y.attrs.copy()
                )
        else:
            return ret

    def get_metadata_sections(self, var: xr.DataArray) -> List[str]:
        # reimplemented to add the Mesh information
        return super().get_metadata_sections(var) + ["Mesh information"]

    get_metadata_sections.__doc__ = (
        psyd.CFDecoder.get_metadata_sections.__doc__
    )

    def get_metadata_for_section(
        self, var: xr.DataArray, section: str, coords: Dict
    ) -> Dict[str, str]:
        # reimplemented to account for the mesh variable
        if section == "Mesh information":
            try:
                mesh = coords[var.attrs["mesh"]]
            except KeyError:
                raise ValueError(
                    f"Could not find mesh variable {var.attrs['mesh']}"
                    f"in the coordinates {tuple(coords)}"
                )
            else:
                return {key: str(val) for key, val in mesh.attrs.items()}
        else:
            return super().get_metadata_for_section(var, section, coords)



#!/usr/bin/env python

# SPDX-FileCopyrightText: 2024 Helmholtz-Zentrum hereon GmbH
#
# SPDX-License-Identifier: LGPL-3.0-only

"""
ugrid classes

set of classes for working with unstructured model grids

The "ugrid" class is the base class: it stores everything in memory

It can read from and write to netcdf files in the UGRID format.

It may be able to reference a netcdf file at some point, rather than storing
directly in memory.

NOTE: only tested for triangular and quad mesh grids at the moment.

NOTE: This code has been extracted from the pull request by @Chilipp to the
gridded package, at https://github.com/NOAA-ORR-ERD/gridded/pull/62/.
Gridded is licensed under the Unlicense.

"""

from __future__ import absolute_import, division, print_function

import numpy as np

# datatype used for indexes -- might want to change for 64 bit some day.
IND_DT = np.int32
NODE_DT = np.float64  # datatype used for node coordinates.


class UGrid(object):
    """
    A basic class to hold an unstructured grid as defined in the UGrid convention.

    The internal structure mirrors the netcdf data standard.
    """

    def __init__(
        self,
        nodes=None,
        node_lon=None,
        node_lat=None,
        faces=None,
        edges=None,
        boundaries=None,
        face_face_connectivity=None,
        face_edge_connectivity=None,
        edge_coordinates=None,
        face_coordinates=None,
        boundary_coordinates=None,
        data=None,
        mesh_name="mesh",
        edge_face_connectivity=None,
        edge_orientation=None,
    ):
        """
        ugrid class -- holds, saves, etc. an unstructured grid

        :param nodes=None : the coordinates of the nodes
        :type nodes: (NX2) array of floats

        :param faces=None : the faces of the grid. Indexes for the nodes array.
        :type faces: (NX3) array of integers

        :param edges=None : the edges of the grid. Indexes for the nodes array.
        :type edges: (NX2) array of integers

        :param boundaries=None: specification of the boundaries are usually a
                                subset of edges where boundary condition
                                information, etc is stored.
                                (NX2) integer array of indexes for the nodes
                                array.
        :type boundaries: numpy array of integers

        :param face_face_connectivity=None: connectivity arrays.
        :param face_edge_connectivity=None: connectivity arrays.

        :param edge_coordinates=None: representative coordinate of the edges.
        :param face_coordinates=None: representative coordinate of the faces.
        :param boundary_coordinates=None: representative coordinate of the
                                          boundaries.

        :param edge_coordinates=None: representative coordinate of the edges
        :type edge_coordinates: (NX2) array of floats

        :param face_coordinates=None: representative coordinate of the faces
                                      (NX2) float array
        :type face_coordinates: (NX2) array of floats


        :param boundary_coordinates=None: representative coordinate of the
                                          boundaries
        :type boundary_coordinates: (NX2) array of floats


        :param data = None: associated variables
        :type data: dict of UVar objects

        :param mesh_name = "mesh": optional name for the mesh
        :type mesh_name: string

        :param edge_face_connectivity=None: optional mapping from edge to
                                            attached face index
        :type edge_face_connectivity: (Nx2) array of ints

        :param edge_orientation=None: the orientation for each edge within the
                                      corresponding face from the
                                      `edge_face_connectivity`. ``1`` means,
                                      the edge has the same orientation in
                                      :attr:`faces` and :attr:`edges`, ``-1``
                                      means the opposite.
        :type edge_orientation: (Nx2) masked array of ints with the same shape
                                      as the `edge_face_connectivity` (i.e.
                                      shape ``(n_edges, 2)``)

        Often this is too much data to pass in as literals -- so usually
        specialized constructors will be used instead (load from file, etc).

        The index variables faces can be a masked array. The mask is
        used for so called flexible meshes. Flexible meshes contain
        cells with varying number of nodes per face.  See the flexible
        mesh section in the convention for further details.
        """

        if (nodes is not None) and (
            (node_lon is not None) or (node_lat is not None)
        ):
            raise TypeError(
                "You need to provide a single nodes array "
                "or node_lon and node_lat"
            )
        if nodes is None:
            if node_lon is not None and node_lat is not None:
                nodes = np.ma.column_stack((node_lon, node_lat))
        self.nodes = nodes
        self.faces = faces
        self.edges = edges
        self.boundaries = boundaries

        self.face_face_connectivity = face_face_connectivity
        self.face_edge_connectivity = face_edge_connectivity

        self.edge_face_connectivity = edge_face_connectivity
        self.edge_orientation = edge_orientation

        self.edge_coordinates = edge_coordinates
        self.face_coordinates = face_coordinates
        self.boundary_coordinates = boundary_coordinates

        self.mesh_name = mesh_name

    @property
    def info(self):
        """
        summary of information about the grid
        """
        msg = ["UGrid object:"]

        msg.append("Number of nodes: %i" % len(self.nodes))
        msg.append(
            "Number of faces: %i with %i vertices per face"
            % (len(self.faces), self.num_vertices)
        )
        if self.boundaries is not None:
            msg.append("Number of boundaries: %i" % len(self.boundaries))

        # if self._data:
        #     msg.append("Variables: " + ", ".join([str(v) for v in self._data.keys()]))
        return "\n".join(msg)

    @property
    def num_vertices(self):
        """
        Maximum number of vertices in a face.

        """
        if self._faces is None:
            return None
        else:
            return self._faces.shape[1]

    @property
    def nodes(self):
        return self._nodes

    @property
    def node_lon(self):
        return self._nodes[:, 0]

    @property
    def node_lat(self):
        return self._nodes[:, 1]

    @nodes.setter  # type: ignore[no-redef]
    def nodes(self, nodes_coords):
        # Room here to do consistency checking, etc.
        # For now -- simply make sure it's a numpy array.
        if nodes_coords is None:
            self.nodes = np.zeros((0, 2), dtype=NODE_DT)
        else:
            self._nodes = np.asanyarray(nodes_coords, dtype=NODE_DT)

    @nodes.deleter  # type: ignore[no-redef]
    def nodes(self):
        # If there are no nodes, there can't be anything else.
        self._nodes = np.zeros((0, 2), dtype=NODE_DT)
        self._edges = None
        self._faces = None
        self._boundaries = None

    @property
    def faces(self):
        return self._faces

    @faces.setter
    def faces(self, faces_indexes):
        # Room here to do consistency checking, etc.
        # For now -- simply make sure it's a numpy array.
        if faces_indexes is not None:
            self._faces = np.asanyarray(faces_indexes, dtype=IND_DT)
        else:
            self._faces = None
            # Other things are no longer valid.
            self._face_face_connectivity = None
            self._face_edge_connectivity = None

    @faces.deleter
    def faces(self):
        self._faces = None
        self._faces = None
        # Other things are no longer valid.
        self._face_face_connectivity = None
        self._face_edge_connectivity = None
        self.edge_coordinates = None

    @property
    def edges(self):
        return self._edges

    @edges.setter
    def edges(self, edges_indexes):
        # Room here to do consistency checking, etc.
        # For now -- simply make sure it's a numpy array.
        if edges_indexes is not None:
            self._edges = np.asanyarray(edges_indexes, dtype=IND_DT)
        else:
            self._edges = None
            self._face_edge_connectivity = None

    @edges.deleter
    def edges(self):
        self._edges = None
        self._face_edge_connectivity = None
        self.edge_coordinates = None

    @property
    def boundaries(self):
        return self._boundaries

    @boundaries.setter
    def boundaries(self, boundaries_indexes):
        # Room here to do consistency checking, etc.
        # For now -- simply make sure it's a numpy array.
        if boundaries_indexes is not None:
            self._boundaries = np.asanyarray(boundaries_indexes, dtype=IND_DT)
        else:
            self._boundaries = None

    @boundaries.deleter
    def boundaries(self):
        self._boundaries = None
        self.boundary_coordinates = None

    @property
    def face_face_connectivity(self):
        return self._face_face_connectivity

    @face_face_connectivity.setter
    def face_face_connectivity(self, face_face_connectivity):
        # Add more checking?
        if face_face_connectivity is not None:
            face_face_connectivity = np.asanyarray(
                face_face_connectivity, dtype=IND_DT
            )
            if face_face_connectivity.shape != (
                len(self.faces),
                self.num_vertices,
            ):
                msg = (
                    "face_face_connectivity must be size " "(num_faces, {})"
                ).format
                raise ValueError(msg(self.num_vertices))
        self._face_face_connectivity = face_face_connectivity

    @face_face_connectivity.deleter
    def face_face_connectivity(self):
        self._face_face_connectivity = None

    @property
    def face_edge_connectivity(self):
        return self._face_edge_connectivity

    @face_edge_connectivity.setter
    def face_edge_connectivity(self, face_edge_connectivity):
        # Add more checking?
        if face_edge_connectivity is not None:
            face_edge_connectivity = np.asanyarray(
                face_edge_connectivity, dtype=IND_DT
            )
            if face_edge_connectivity.shape != (
                len(self.faces),
                self.num_vertices,
            ):
                msg = (
                    "face_face_connectivity must be size " "(num_face, {})"
                ).format
                raise ValueError(msg(self.num_vertices))
        self._face_edge_connectivity = face_edge_connectivity

    @face_edge_connectivity.deleter
    def face_edge_connectivity(self):
        self._face_edge_connectivity = None

    def infer_location(self, data, axis=-1):
        """
        :param data:
        :param axis:

        :returns: 'nodes' if data will fit to the nodes,
                  'faces' if the data will fit to the faces,
                  'boundaries' if the data will fit the boundaries.
                  None otherwise.

        If data is a netcdf variable, the "location" attribute is checked.
        """
        # We should never be calling infer_locations if it was already defined
        # try:
        #     loc = data.location
        #     if loc == "face":
        #         # FIXME: should we check the array size in this case?
        #         return "face"
        # except AttributeError:
        #     pass # try checking array size
        # # fixme: should use UGRID compliant nc_attributes if possible
        try:
            size = data.shape[axis]
        except IndexError:
            return None  # Variable has a size-zero data array
        if size == self.nodes.shape[0]:
            return "node"
        if self.faces is not None and size == self.faces.shape[0]:
            return "face"
        if self.boundaries is not None and size == self.boundaries.shape[0]:
            return "boundary"
        return None

    def build_face_face_connectivity(self):
        """
        Builds the face_face_connectivity array: giving the neighbors of each cell.

        Note: arbitrary order and CW vs CCW may not be consistent.
        """

        num_vertices = self.num_vertices
        num_faces = self.faces.shape[0]
        face_face = np.zeros((num_faces, num_vertices), dtype=IND_DT)
        face_face += -1  # Fill with -1.

        # Loop through all the faces to find the matching edges:
        edges = {}  # dict to store the edges.
        for i, face in enumerate(self.faces):
            # Loop through edges of the cell:
            for j in range(num_vertices):
                if j < self.num_vertices - 1:
                    edge = (face[j], face[j + 1])
                else:
                    edge = (face[-1], face[0])
                if edge[0] > edge[1]:  # Sort the node numbers.
                    edge = (edge[1], edge[0])
                # see if it is already in there
                prev_edge = edges.pop(edge, None)
                if prev_edge is not None:
                    face_num, edge_num = prev_edge
                    face_face[i, j] = face_num
                    face_face[face_num, edge_num] = i
                else:
                    edges[edge] = (i, j)  # face num, edge_num.
        self._face_face_connectivity = face_face

    def get_lines(self):
        if self.edges is None:
            self.build_edges()
        return self.nodes[self.edges]

    def build_edges(self):
        """
        Builds the edges array: all the edges defined by the faces

        This will replace the existing edge array, if there is one.

        NOTE: arbitrary order -- should the order be preserved?
        """
        if self.faces is None:
            # No faces means no edges
            self._edges = None
            return

        faces = self.faces

        is_masked = np.ma.isMA(faces)
        if is_masked:
            first = faces.copy()
            first[:] = faces[:, :1]
            save_mask = faces.mask.copy()
            faces[save_mask] = first.data[faces.mask]

        face_edges = np.dstack([faces, np.roll(faces, 1, 1)])

        if is_masked and np.ndim(save_mask):
            face_edges.mask = np.dstack(
                [np.zeros_like(save_mask), np.roll(save_mask, 1, 1)]
            )

        face_edges.sort(axis=-1)

        all_edges = face_edges.reshape((-1, 2))

        if is_masked and np.ndim(save_mask):
            edges = np.unique(all_edges[~all_edges.mask.any(axis=-1)], axis=0)
        else:
            edges = np.unique(all_edges, axis=0)
        self._edges = edges

    def build_boundaries(self):
        """
        Builds the boundary segments from the cell array.

        It is assumed that -1 means no neighbor, which indicates a boundary

        This will over-write the existing boundaries array if there is one.

        This is a not-very-smart just loop through all the faces method.

        """
        boundaries = []
        for i, face in enumerate(self.face_face_connectivity):
            for j, neighbor in enumerate(face):
                if neighbor == -1:
                    if j == self.num_vertices - 1:
                        bound = (self.faces[i, -1], self.faces[i, 0])
                    else:
                        bound = (self.faces[i, j], self.faces[i, j + 1])
                    boundaries.append(bound)
        self.boundaries = boundaries

    def build_face_edge_connectivity(self):
        """
        Builds the face-edge connectivity array
        """
        self.face_edge_connectivity = self._build_face_edge_connectivity()

    def _build_face_edge_connectivity(self, sort=True):
        try:
            from scipy.spatial import cKDTree
        except ImportError:
            raise ImportError(
                "The scipy package is required to use "
                "UGrid.build_face_edge_connectivity"
            )

        faces = self.faces.copy()
        if self.edges is None:
            self.build_edges()
        edges = self.edges.copy()

        is_masked = np.ma.isMA(faces)
        if is_masked:
            first = faces.copy()
            first[:] = faces[:, :1]
            save_mask = faces.mask.copy()
            faces[save_mask] = first.data[faces.mask]

        face_edges = np.dstack([faces, np.roll(faces, 1, 1)])

        if is_masked and np.ndim(save_mask):
            face_edges.mask = np.dstack(
                [np.zeros_like(save_mask), np.roll(save_mask, 1, 1)]
            )

        if sort:
            face_edges.sort(axis=-1)
            edges.sort(axis=-1)

        tree = cKDTree(edges)

        face_edge_2d = face_edges.reshape((-1, 2))

        if is_masked and save_mask.any():
            mask = face_edge_2d.mask.any(-1)
            connectivity = np.ma.ones(
                len(face_edge_2d),
                dtype=face_edge_2d.dtype,
            )
            connectivity.mask = mask
            connectivity[~mask] = tree.query(
                face_edge_2d[~mask], distance_upper_bound=0.1
            )[1]
        else:
            connectivity = tree.query(face_edge_2d, distance_upper_bound=0.1)[
                1
            ]
        return np.roll(connectivity.reshape(faces.shape), -1, -1)

    def get_face_edge_orientation(self):
        """
        Get the orientation for each edge in the corresponding face

        This method returns an array with the same shape as :attr:`faces` that
        is one if the corresponding edge has the same orientation as in
        :attr:`edges`, and -1 otherwise
        """
        # we build the face edge connectivity but do not sort the edge nodes.
        # With this, we will get `num_edges` where the edge is flipped compared
        # to the definition in :attr:`edges`
        face_edge_connectivity = self._build_face_edge_connectivity(sort=False)
        num_edges = self.edges.shape[0]
        if np.ma.isMA(face_edge_connectivity):
            return np.ma.where(face_edge_connectivity == num_edges, 1, -1)
        else:
            return np.where(face_edge_connectivity == num_edges, 1, -1)

    def build_edge_face_connectivity(self):
        """Build the edge_face_connectivity

        The edge_face_connectivity is the mapping from each edge in the
        :attr:`edges` to the attached face in `faces`.
        """
        if self.face_edge_connectivity is None:
            self.build_face_edge_connectivity()
        face_edge_connectivity = self.face_edge_connectivity
        orientation = self.get_face_edge_orientation()

        n_edge = fill_value = len(self.edges)
        n_face = len(self.faces)

        if np.ma.isMA(face_edge_connectivity):
            face_edge_connectivity = face_edge_connectivity.filled(fill_value)

        n_face, nmax_edge = face_edge_connectivity.shape
        # Get rid of the fill_value, create a 1:1 mapping between faces and edges
        isnode = (face_edge_connectivity != fill_value).ravel()
        face_index = np.repeat(np.arange(n_face), nmax_edge).ravel()[isnode]
        orientation_nodes = orientation.ravel()[isnode]
        edge_index = face_edge_connectivity.ravel()[isnode]

        # We know that every edge will have either one or two associated faces
        isface = np.empty((n_edge, 2), dtype=bool)
        isface[:, 0] = True
        isface[:, 1] = np.bincount(edge_index) == 2

        # Allocate the output array
        edge_face_connectivity = np.full((n_edge, 2), n_face, dtype=np.int64)
        # Invert the face_index, and use the boolean array to place them appropriately
        edge_face_connectivity.ravel()[isface.ravel()] = face_index[
            np.argsort(edge_index)
        ]
        self.edge_face_connectivity = np.ma.masked_where(
            edge_face_connectivity == n_face, edge_face_connectivity
        )

        edge_orientation = np.full((n_edge, 2), -999, dtype=np.int64)
        # Invert the face_index, and use the boolean array to place them appropriately
        edge_orientation.ravel()[isface.ravel()] = orientation_nodes[
            np.argsort(edge_index)
        ]
        self.edge_orientation = np.ma.masked_where(
            edge_orientation == -999, edge_orientation
        )

    def _get_node_edge_connectivity_unsorted(self):
        """Build the node_edge_connectivity.

        The node_edge_connectivity is the mapping from each node in the
        :attr:`nodes` to the attached edge in :attr:`edges`. Note that this
        method does not sort the edges so they are in general not in
        anti-clockwise order.
        """
        if self.edges is None:
            self.build_edges()
        edge_node_connectivity = self.edges

        n_edge = len(self.edges)
        n_edge, nmax_node = edge_node_connectivity.shape
        n_node = fill_value = len(self.nodes)

        if np.ma.isMA(edge_node_connectivity):
            edge_node_connectivity = edge_node_connectivity.filled(fill_value)

        # Get rid of the fill_value, create a 1:1 mapping between edges and
        # nodes
        isnode = (edge_node_connectivity != fill_value).ravel()
        edge_index = np.repeat(np.arange(n_edge), nmax_node).ravel()[isnode]
        node_index = edge_node_connectivity.ravel()[isnode]

        node_counts = np.bincount(node_index)
        nmax_edge = node_counts.max()

        # We know that every edge will have either one or two associated faces
        isedge = np.empty((n_node, nmax_edge), dtype=bool)
        for i in range(nmax_edge):
            isedge[:, i] = node_counts > i

        # Allocate the output array
        node_edge_connectivity = np.full(
            (n_node, nmax_edge), n_edge, dtype=np.int64
        )
        # Invert the face_index, and use the boolean array to place them
        # appropriately
        node_edge_connectivity.ravel()[isedge.ravel()] = edge_index[
            np.argsort(node_index)
        ]
        return np.ma.masked_where(
            node_edge_connectivity == n_edge, node_edge_connectivity
        )

    def _create_dual_edge_mesh(self):
        """Create a :class:`UGrid` instance that represents the dual edge mesh."""
        if self.face_edge_connectivity is None:
            self.build_face_edge_connectivity()

        edges = self.edges

        if self.edge_face_connectivity is None:
            self.build_edge_face_connectivity()

        n_face = len(self.faces)
        n_node = len(self.nodes)

        edge_face_connectivity = self.edge_face_connectivity.filled(n_face)

        # now get the orientation for each edge from the `orientation` array
        mask = edge_face_connectivity < n_face
        edge_orientation = self.edge_orientation.filled(-999)

        # use num_faces as fill value (necessary for edges at the domain boundary)
        dual_face_node_connectivity = np.full(
            (len(edges), 4), -999, dtype=self.edges.dtype
        )
        dual_face_node_connectivity[:, 0] = edges[:, 0]
        dual_face_node_connectivity[:, 2] = edges[:, 1]

        # get the first index for the face center nodes
        if self.face_coordinates is None:
            self.build_face_coordinates()

        dual_nodes = np.r_[self.nodes, self.face_coordinates]

        # now handle the case where the orientation is -1. This should be at
        # dual_face_node_connectivity[:, 1]
        mask = edge_orientation == -1
        dual_face_node_connectivity[mask.any(axis=-1), 3] = (
            edge_face_connectivity[mask] + n_node
        )

        # the same for +1, should be at dual_face_node_connectivity[:, 3]
        mask = edge_orientation == 1
        dual_face_node_connectivity[mask.any(axis=-1), 1] = (
            edge_face_connectivity[mask] + n_node
        )

        # now we need to roll where dual_face_node_connectivity[:, 1] == -999
        # to make sure that the fill values are at the end
        roll_at = dual_face_node_connectivity[:, 1] == -999
        dual_face_node_connectivity[roll_at] = np.roll(
            dual_face_node_connectivity[roll_at], 2, axis=1
        )

        # now turn dual_face_node_connectivity into a masked array
        # NOTE: There is no definititive policy yet how to deal with fill
        # values within the gridded package, see
        # https://github.com/NOAA-ORR-ERD/gridded/pull/60#issuecomment-744810919
        dual_face_node_connectivity = np.ma.masked_where(
            dual_face_node_connectivity == -999, dual_face_node_connectivity
        )

        return dual_face_node_connectivity.astype(int), dual_nodes

    def _create_dual_node_mesh(self):
        """Create the dual mesh for the nodes."""
        from psy_ugrid._create_dual_node_mesh import get_face_node_connectivity

        (
            dual_edge_face_node_connectivity,
            dual_nodes,
        ) = self._create_dual_edge_mesh()

        # create a node_edge_connectivty
        node_edge_connectivity = self._get_node_edge_connectivity_unsorted()

        if self.edge_coordinates is None:
            self.build_edge_coordinates()

        edge_coordinates = self.edge_coordinates

        n_edge = len(self.edges)
        n_node = len(self.nodes)
        n_dual_node = len(dual_nodes)
        n_dual_node_max = n_dual_node + n_edge

        face_node_connectivity = self.faces
        if np.ma.isMA(face_node_connectivity):
            face_node_connectivity = face_node_connectivity.filled(
                len(self.nodes)
            )
        nmax_face = (
            np.bincount(
                face_node_connectivity[face_node_connectivity < n_node]
            ).max()
            + 3
        )

        node_edge_connectivity = node_edge_connectivity.filled(n_edge)
        dual_edge_face_node_connectivity = (
            dual_edge_face_node_connectivity.filled(n_dual_node_max)
        )

        dual_node_face_node_connectivity = np.full(
            (n_node, nmax_face), int(n_dual_node_max), dtype=np.int64
        )

        dual_node_face_node_connectivity = np.asarray(
            get_face_node_connectivity(
                dual_edge_face_node_connectivity,
                node_edge_connectivity,
                n_dual_node,
                nmax_face,
            )
        )

        is_new_node = dual_node_face_node_connectivity >= n_dual_node
        all_new = dual_node_face_node_connectivity[is_new_node]
        new_nodes = np.unique(dual_node_face_node_connectivity[is_new_node])

        dual_node_face_node_connectivity[
            is_new_node
        ] = n_dual_node + new_nodes.searchsorted(all_new)
        n_dual_node_max = n_dual_node + len(new_nodes) - 1

        return (
            np.ma.masked_where(
                dual_node_face_node_connectivity == n_dual_node_max,
                dual_node_face_node_connectivity,
            ),
            np.r_[dual_nodes, edge_coordinates[new_nodes[:-1] - n_dual_node]],
        )

    def create_dual_mesh(self, location="edge"):
        """Create the dual mesh for edge or nodes.

        This method creates the dual mesh, either specified through the nodes,
        or specified through the edges. For a Delaunay triangulation case with
        ``location == "node"``, this is commonly known as Voronoi Polygons.

        :param location="edge" : the source for the dual mash. can be one of
                                 ``"node"`` or ``"edge"``
        :type location: str

        :returns: A :class:`UGrid` with `nodes` and `faces` of the dual mesh.
        """
        if location == "edge":
            face_node_connectivity, nodes = self._create_dual_edge_mesh()
        elif location == "node":
            face_node_connectivity, nodes = self._create_dual_node_mesh()
        else:
            raise ValueError(
                "location must be `edge` or `node`, found `%s`" % (location,)
            )
        if self.mesh_name:
            mesh_name = self.mesh_name + "_dual_" + location
        else:
            mesh_name = "dual_" + location
        return UGrid(nodes, faces=face_node_connectivity, mesh_name=mesh_name)

    def build_face_coordinates(self):
        """
        Builds the face_coordinates array, using the average of the
        nodes defining each face.

        Note that you may want a different definition of the face
        coordinates than this computes, but this is here to have
        an easy default.

        This will write-over an existing face_coordinates array.

        Useful if you want this in the output file.

        """
        faces = self.faces
        if not np.ma.isMA(faces) or not np.ndim(faces.mask):
            self.face_coordinates = self.nodes[faces].mean(axis=1)
        else:
            mask = np.dstack([faces.mask, faces.mask])
            coords = self.nodes[faces.filled(0)]
            coords[mask] = np.nan
            self.face_coordinates = np.nanmean(coords, axis=1)

    def build_edge_coordinates(self):
        """
        Builds the edge_coordinates array, using the average of the
        nodes defining each edge.

        Note that you may want a different definition of the edge
        coordinates than this computes, but this is here to have
        an easy default.


        This will write-over an existing edge_coordinates array

        Useful if you want this in the output file

        """
        self.edge_coordinates = self.nodes[self.edges].mean(axis=1)

    def build_boundary_coordinates(self):
        """
        Builds the boundary_coordinates array, using the average of the
        nodes defining each boundary segment.

        Note that you may want a different definition of the boundary
        coordinates than this computes, but this is here to have
        an easy default.

        This will write-over an existing face_coordinates array

        Useful if you want this in the output file

        """
        boundary_coordinates = np.zeros(
            (len(self.boundaries), 2), dtype=NODE_DT
        )
        # FXIME: there has got to be a way to vectorize this.
        for i, bound in enumerate(self.boundaries):
            coords = self.nodes[bound]
            boundary_coordinates[i] = coords.mean(axis=0)
        self.boundary_coordinates = boundary_coordinates



# SPDX-FileCopyrightText: 2024 Helmholtz-Zentrum hereon GmbH
#
# SPDX-License-Identifier: LGPL-3.0-only

"""Test file for building the face_edge_connectivity"""
import numpy as np

from psy_ugrid import ugrid


def test_build_face_edge_connectivity():
    faces = [[0, 1, 2], [1, 2, 3]]
    edges = [[0, 1], [1, 2], [2, 0], [2, 3], [3, 1]]
    nodes = [1, 2, 3, 4]
    grid = ugrid.UGrid(
        node_lon=nodes, node_lat=nodes, faces=faces, edges=edges
    )

    ref = [[0, 1, 2], [1, 3, 4]]
    grid.build_face_edge_connectivity()

    assert grid.face_edge_connectivity.tolist() == ref


def test_build_face_edge_connectivity_na():
    faces = np.ma.array(
        [[0, 1, 2, 3], [1, 2, 3, -999]],
        mask=[[False, False, False, False], [False, False, False, True]],
    )
    edges = [[0, 1], [1, 2], [2, 3], [3, 0], [3, 1]]
    nodes = [1, 2, 3, 4]
    grid = ugrid.UGrid(
        node_lon=nodes, node_lat=nodes, faces=faces, edges=edges
    )

    ref = [[0, 1, 2, 3], [1, 2, 4, -999]]
    grid.build_face_edge_connectivity()

    assert grid.face_edge_connectivity.filled(-999).tolist() == ref



# SPDX-FileCopyrightText: 2024 Helmholtz-Zentrum hereon GmbH
#
# SPDX-License-Identifier: LGPL-3.0-only

"""Test file for building the face_edge_connectivity"""
import numpy as np

from psy_ugrid import ugrid


def test_get_face_edge_orientation():
    faces = [[0, 1, 2], [1, 2, 3]]
    edges = [[0, 1], [1, 2], [0, 2], [3, 2], [3, 1]]
    nodes = [1, 2, 3, 4]
    grid = ugrid.UGrid(
        node_lon=nodes, node_lat=nodes, faces=faces, edges=edges
    )

    ref = [[1, 1, -1], [1, -1, 1]]
    orientation = grid.get_face_edge_orientation()

    assert orientation.tolist() == ref


def test_get_face_edge_orientation_na():
    faces = np.ma.array(
        [[0, 1, 2, 3], [1, 2, 3, -999]],
        mask=[[False, False, False, False], [False, False, False, True]],
    )
    edges = [[0, 1], [1, 2], [2, 3], [0, 3], [1, 3]]
    nodes = [1, 2, 3, 4]
    grid = ugrid.UGrid(
        node_lon=nodes, node_lat=nodes, faces=faces, edges=edges
    )

    ref = [[1, 1, 1, -1], [1, 1, -1, -999]]
    orientation = grid.get_face_edge_orientation()

    assert orientation.filled(-999).tolist() == ref



# SPDX-FileCopyrightText: 2024 Helmholtz-Zentrum hereon GmbH
#
# SPDX-License-Identifier: LGPL-3.0-only

"""psy-ugrid

A psyplot plugin for decoding unstructured grids following the UGRID conventions
"""

from __future__ import annotations

from . import _version

__version__ = _version.get_versions()["version"]

__author__ = "Philipp S. Sommer"
__copyright__ = "2024 Helmholtz-Zentrum hereon GmbH"
__credits__ = [
    "Philipp S. Sommer",
]
__license__ = "LGPL-3.0-only"

__maintainer__ = "Philipp S. Sommer"
__email__ = "philipp.sommer@hereon.de"

__status__ = "Pre-Alpha"



"""Utility function for creating the dual mesh for the nodes."""

# SPDX-FileCopyrightText: 2024 Helmholtz-Zentrum hereon GmbH
#
# SPDX-License-Identifier: LGPL-3.0-only

# isort: off
from cython cimport view
# isort: on


def get_face_node_connectivity(
    const long[:, :] dual_edge_face_node_connectivity,
    const long[:, :] node_edge_connectivity,
    long n_dual_node,
    long nmax_face,
):
    """Create the face_node_connectivity for the dual node mesh.

    This function loops through every node and creates the corresponding
    elements.
    """
    cdef n_node = len(node_edge_connectivity)
    cdef n_edge = len(dual_edge_face_node_connectivity)

    cdef long[:, :] ret = view.array(
        shape=(n_node, nmax_face), itemsize=sizeof(long), format="l"
    )

    cdef long n_dual_node_max = n_dual_node + n_edge
    cdef const long[:] edges

    cdef long i, j, node, i_edge
    cdef long[:] edge
    cdef long nmax_edge = node_edge_connectivity.shape[1]
    cdef long[:, :] dual_cells = view.array(
        shape=(nmax_edge, 4), itemsize=sizeof(long), format="l"
    )
    cdef long[:, :] sorted_edges = view.array(
        shape=(nmax_edge + 2, 2), itemsize=sizeof(long), format="l"
    )
    cdef long[:, :] dual_edges = view.array(
        shape=(nmax_edge + 2, 2), itemsize=sizeof(long), format="l"
    )
    cdef long[:] dual_cell = view.array(
        shape=(4, ), itemsize=sizeof(long), format="l"
    )

    cdef long n_cell_edges = nmax_edge
    cdef long start, end

    ret[:, :] = n_dual_node_max

    for node in range(n_node):
        edges = node_edge_connectivity[node]

        # clear the arrays
        dual_cells[:, :] = n_dual_node_max
        dual_edges[:, :] = n_dual_node_max
        sorted_edges[:, :] = n_dual_node_max
        if edges[0] == n_edge:
            continue

        for i in range(nmax_edge):
            i_edge = edges[i]
            if i_edge < n_edge:
                dual_cell[:] = dual_edge_face_node_connectivity[i_edge]
                n_cell_edges = i + 1

                # make sure that the node is in the first position
                if dual_cell[0] != node:
                    dual_cell[:2] = dual_cell[2:4]
                    dual_cell[2:4] = dual_edge_face_node_connectivity[i_edge, :2]

                dual_cells[i, :] = dual_cell

        dual_edges[:nmax_edge] = dual_cells[:, 1::2]

        # now check for missing nodes (indicates that the node is at the
        # edge of the mesh)
        for i in range(n_cell_edges):
            i_edge = edges[i]
            edge = dual_edges[i]
            if edge[0] == n_dual_node_max:
                edge[0] = n_dual_node + i_edge
                dual_edges[n_cell_edges, 0] = node
                dual_edges[n_cell_edges, 1] = n_dual_node + i_edge
                n_cell_edges = n_cell_edges + 1
            elif edge[1] == n_dual_node_max:
                edge[1] = n_dual_node + i_edge
                dual_edges[n_cell_edges, 1] = node
                dual_edges[n_cell_edges, 0] = n_dual_node + i_edge
                n_cell_edges = n_cell_edges + 1

        # now sort the edges
        sorted_edges[0, :] = dual_edges[0]
        for i in range(1, n_cell_edges):
            end = sorted_edges[i - 1, 1]
            for j in range(1, n_cell_edges):
                start = dual_edges[j, 0]
                if start == end:
                    sorted_edges[i] = dual_edges[j]
                    break
        ret[node, :n_cell_edges] = sorted_edges[:n_cell_edges, 0]

    return ret



# SPDX-FileCopyrightText: 2024 Helmholtz-Zentrum hereon GmbH
#
# SPDX-License-Identifier: LGPL-3.0-only

import psyplot.data as psyd

from psy_ugrid.decoder import UGridDecoder


def test_get_decoder(get_test_file):
    """Test to get the right decoder"""
    ds = psyd.open_dataset(get_test_file("simple_triangular_grid_si0.nc"))
    d = psyd.CFDecoder.get_decoder(ds, ds.Mesh2_fcvar)
    assert isinstance(d, UGridDecoder)


def test_x(get_test_file):
    """Test the get_x method"""
    ds = psyd.open_dataset(get_test_file("simple_triangular_grid_si0.nc"))
    d = psyd.CFDecoder.get_decoder(ds, ds.Mesh2_fcvar)

    x = d.get_x(ds.Mesh2_fcvar)
    assert "standard_name" in x.attrs
    assert x.attrs["standard_name"] == "longitude"
    rounded = list(map(lambda v: round(v, 3), x.values))
    assert rounded == [0.3, 0.567]


def test_y(get_test_file):
    """Test the get_y method"""
    ds = psyd.open_dataset(get_test_file("simple_triangular_grid_si0.nc"))
    d = psyd.CFDecoder.get_decoder(ds, ds.Mesh2_fcvar)

    y = d.get_y(ds.Mesh2_fcvar)
    assert "standard_name" in y.attrs
    assert y.attrs["standard_name"] == "latitude"
    rounded = list(map(lambda v: round(v, 3), y.values))
    assert rounded == [0.4, 0.767]



# SPDX-FileCopyrightText: 2024 Helmholtz-Zentrum hereon GmbH
#
# SPDX-License-Identifier: LGPL-3.0-only

# coding: utf-8
"""Test file for building the face_edge_connectivity"""
import numpy as np

from psy_ugrid import ugrid


def test_create_dual_edge_mesh():
    r"""Test for a mesh like

     /|\
    /_|_\
    """
    nodex = [0, 1, 1, 2]
    nodey = [0, 1, 0, 0]
    faces = [[0, 1, 2], [2, 1, 3]]

    edges = [[0, 1], [1, 2], [2, 0], [1, 3], [2, 3]]
    grid = ugrid.UGrid(
        node_lon=nodex, node_lat=nodey, faces=faces, edges=edges
    )

    center1 = len(nodex)
    center2 = center1 + 1
    ref = [
        [0, center1, 1, -999],
        [1, center1, 2, center2],
        [2, center1, 0, -999],
        [1, center2, 3, -999],
        [3, center2, 2, -999],
    ]

    dual_faces = grid._create_dual_edge_mesh()[0]

    assert np.ma.isMA(dual_faces)
    assert dual_faces.filled(-999).tolist() == ref


def test_create_dual_edge_mesh_na():
    r"""Test for a mesh like

    |‾|\
    |_|_\
    """
    nodex = [0, 0, 1, 1, 2]
    nodey = [0, 1, 1, 0, 0]
    faces = np.ma.array(
        [[0, 1, 2, 3], [2, 4, 3, -999]],
        mask=[[False, False, False, False], [False, False, False, True]],
    )

    edges = [[0, 1], [1, 2], [2, 3], [3, 0], [2, 4], [4, 3]]
    grid = ugrid.UGrid(
        node_lon=nodex, node_lat=nodey, faces=faces, edges=edges
    )

    center1 = len(nodex)
    center2 = center1 + 1
    ref = [
        [0, center1, 1, -999],
        [1, center1, 2, -999],
        [2, center1, 3, center2],
        [3, center1, 0, -999],
        [2, center2, 4, -999],
        [4, center2, 3, -999],
    ]

    dual_faces = grid._create_dual_edge_mesh()[0]
    assert np.ma.isMA(dual_faces)
    assert dual_faces.filled(-999).tolist() == ref



# SPDX-FileCopyrightText: 2024 Helmholtz-Zentrum hereon GmbH
#
# SPDX-License-Identifier: LGPL-3.0-only

"""Test file for imports."""


def test_package_import():
    """Test the import of the main package."""
    import psy_ugrid  # noqa: F401
    import psy_ugrid._create_dual_node_mesh  # noqa: F401



# SPDX-FileCopyrightText: 2024 Helmholtz-Zentrum hereon GmbH
#
# SPDX-License-Identifier: LGPL-3.0-only

"""psy-ugrid psyplot plugin

This module defines the rcParams for the psy-ugrid plugin. This module will
be imported when psyplot is imported. What is should contain is:

- an rcParams variable as instance of :class:`psyplot.config.rcsetup.RcParams`
  that describes the configuration of your plugin
- a get_versions function that returns the version of your plugin and the ones
  from its requirements

.. warning::

    Because of recursion issues, You have to load the psyplot module before
    loading this module! In other words, you have to type

    .. code-block:: python

        import psyplot
        import psy_ugrid.plugin
"""
from __future__ import annotations

from typing import Any, Dict

import psyplot.data as psyd
from psyplot.config.rcsetup import RcParams

from psy_ugrid import __version__ as plugin_version
from psy_ugrid.decoder import UGridDecoder


def get_versions(requirements=True):
    """Get the versions of psy-ugrid and it's requirements

    Parameters
    ----------
    requirements: bool
        If True, the requirements are imported and it's versions are included
    """
    ret = {"version": plugin_version}
    if requirements:
        # insert versions of the requirements, e.g. via::
        #
        #     import requirement
        #     ret['requirement'] = requirement.__version__
        #
        pass
    return ret


# -----------------------------------------------------------------------------
# ------------------------- validation functions ------------------------------
# -----------------------------------------------------------------------------


# define your validation functions for the values in the rcParams here. If
# a validation fails, the function should raise a ValueError or TypeError


# -----------------------------------------------------------------------------
# ------------------------------ rcParams -------------------------------------
# -----------------------------------------------------------------------------


# define your defaultParams. A mapping from rcParams key to a list of length 3:
#
# 1. the default value
# 2. the validation function of type conversion function
# 3. a short description of the default value
#
# Example::
#
#     defaultParams = {'my.key': [True, bool, 'What my key does']}
defaultParams: Dict[str, Any] = {}

# create the rcParams and populate them with the defaultParams. For more
# information on this class, see the :class:`psyplot.config.rcsetup.RcParams`
# class
rcParams = RcParams(defaultParams=defaultParams)
rcParams.update_from_defaultParams()


psyd.CFDecoder.register_decoder(UGridDecoder)



# SPDX-FileCopyrightText: 2024 Helmholtz-Zentrum hereon GmbH
#
# SPDX-License-Identifier: CC0-1.0

"""Setup script for the psy-ugrid package."""
import os

import versioneer
from setuptools import Extension, setup

_USE_CYTHON = os.getenv("USE_CYTHON", "").lower()
# use cython, when there is no c code or when it is specified via environment
# variable
USE_CYTHON = not os.path.exists(
    os.path.join("psy_ugrid", "_create_dual_node_mesh.c")
) or (
    bool(_USE_CYTHON)
    and (not _USE_CYTHON.startswith("f") and not _USE_CYTHON.startswith("n"))
)

ext = ".pyx" if USE_CYTHON else ".c"


extensions = [
    Extension(
        "psy_ugrid._create_dual_node_mesh",
        sources=[os.path.join("psy_ugrid", "_create_dual_node_mesh" + ext)],
    )
]


if USE_CYTHON:
    from Cython.Build import cythonize

    extensions = cythonize(extensions)


setup(
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    ext_modules=extensions,
)
